from cudagrad.nn import Module as Module
from cudagrad.nn import mse as mse
from cudagrad.nn import sgd as sgd
from cudagrad.tensor import Tensor as Tensor
from cudagrad.tensor import hello as hello

__version__ = "0.2.4"



# So the magic is that there's this relatively simple algorithm called
# backpropagation that takes the error in the output and sends that
# error backwards through the network and computes through all the
# connections how you should change them to improve the behavior, and
# then you change them all a tiny bit and you just keep going with
# another example. And surprisingly that actually works. For many years
# people thought that would just get jammed up — it would get stuck
# somewhere — but no it doesn't, it actually works very well.
#
# Geoffrey Hinton

from cudagrad.tensor import Tensor


def mse(predicted: Tensor, actual: Tensor) -> Tensor:
    # TODO this assumes column tensor
    difference = predicted - actual
    squared = difference * difference
    summed = squared.sum()
    mean = summed / Tensor([1], [summed.size[0]])
    return mean


class Module:
    def parameters(self) -> list[Tensor]:
        params = []
        for x in self.__dict__:
            param = getattr(self, x)
            if type(param) is Tensor:
                params.append(param)
        return params

    def zero_grad(self) -> None:
        for parameter in self.parameters():
            assert type(parameter) is Tensor
            parameter.zero_grad()


def sgd(model: Module, lr: float) -> None:
    for parameter in model.parameters():
        parameter.data[:] = parameter + (
            Tensor.explode(parameter.size, -lr) * parameter.grad()
        )


def softmax(t: Tensor) -> Tensor:
    D = Tensor([1], [-1]) * t.max()
    e = (t + D).exp()
    return e / e.sum()



import os
import subprocess
import sys
from shutil import move

import toml
from setuptools import Extension, find_packages, setup

from pybind11.setup_helpers import build_ext


class CMakeExtension(Extension):
    def __init__(self, name: str, sourcedir: str = "") -> None:
        Extension.__init__(self, name, sources=[])
        self.sourcedir = os.path.abspath(sourcedir)


class CMakeBuild(build_ext):
    def run(self) -> None:
        try:
            subprocess.check_output(["cmake", "--version"])
        except OSError:
            raise RuntimeError(
                "CMake must be installed to build the following extensions: "
                + ", ".join(e.name for e in self.extensions)
            )

        for ext in self.extensions:
            self.build_extension(ext)

    def build_extension(self, ext: CMakeExtension) -> None:
        extdir = os.path.abspath(os.path.dirname(self.get_ext_fullpath(ext.name)))
        cmake_args = [
            "-DCMAKE_LIBRARY_OUTPUT_DIRECTORY=" + extdir,
            "-DPYTHON_EXECUTABLE=" + sys.executable,
            "-DPython3_EXECUTABLE=" + sys.executable,
        ]

        cfg = "Debug" if self.debug else "Release"
        build_args = ["--config", cfg]

        if not os.path.exists(self.build_temp):
            os.makedirs(self.build_temp)

        subprocess.check_call(
            ["cmake", ext.sourcedir] + cmake_args, cwd=self.build_temp
        )
        subprocess.check_call(
            ["cmake", "--build", ".", "--target", ext.name] + build_args,
            cwd=self.build_temp,
        )
        move(
            os.path.join(self.build_lib, "tensor.so"),
            os.path.join(extdir, "cudagrad", "tensor.so"),
        )


def get_version_from_toml() -> str:
    data = toml.load("pyproject.toml")
    version = str(data.get("project", {}).get("version", None))
    if version is None:
        raise RuntimeError("Can't get version in TOML!")
    else:
        return version


__version__ = get_version_from_toml()

setup(
    name="cudagrad",
    version=__version__,
    author="Ryan Moore",
    author_email="ryanm.inbox@gmail.com",
    url="https://github.com/yrmo/cudagrad",
    description="A tensor-valued autograd engine for Python",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    ext_modules=[CMakeExtension("tensor")],
    cmdclass=dict(build_ext=CMakeBuild),
    zip_safe=False,
    python_requires=">=3.7",
    packages=find_packages(),
    include_package_data=True,
)
